/*******************************************************************************
**	Cirrus Logic Laguna 546x chipset driver
**      Text 16 functions
********************************************************************************
**
   Copyright (C) 1997 Rodolphe Ortalo, [ortalo@laas.fr]
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
**
********************************************************************************
**	$Log: 546x_text16.inc,v $
**	Revision 1.5  1998/11/19 02:43:03  marcus
**	License change
**	
**	Revision 1.4  1998/10/25 18:09:42  marcus
**	More namespace cleanups
**	
**	Revision 1.3  1998/10/25 02:04:31  marcus
**	Namespace cleanup to make KGI/LibGGI interaction possible
**	
**	Revision 1.2  1998/08/25 12:11:51  marcus
**	*** empty log message ***
**	
**	Revision 1.1  1998/07/07 07:22:30  taylorj
**	*** empty log message ***
**	
**	Revision 1.7  1998/02/17 00:06:14  ortalo
**	Incorporated the text<->dpp changes (done by MenTalBoy).
**	
**	Revision 1.6  1998/01/27 23:46:11  ortalo
**	Partial support of 2D bitBLT hardware accelerations
**	(tiling management, drawbox in all modes).
**	
**	Revision 1.5  1997/12/14 21:55:35  becka
**	uint/sint namespace fixed ... please test ...
**	
**	Revision 1.4  1997/11/27 22:22:45  ortalo
**	Changed all the old long macro definitions to shorter ones. (Register
**	names now fit my own coding style.)
**	Several clean-ups in chipset code, but no functional improvements.
**	Still some cleaning to do.
**	
**	Revision 1.3  1997/09/15 21:52:13  ortalo
**	Added set_font functions (adapted from the vga driver). Text mode restore
**	after a graphics mode set now works (thanks to all).
**	
**	Revision 1.2  1997/09/14 19:05:11  ortalo
**	Graphic mode setting and frame buffer MMIO access works (incorrectly
**	but... well...). The return to text mode after a graphic mode set
**	does not work ! (Either a bug, or the absence of set font functions).
**	
**	Revision 1.1  1997/08/31 18:57:14  ortalo
**	Text 9x16 mode setting works. Added a separate file for text functions.
**	Added set_origin/split functions.
**	
*/

/*
** Font setting
*/

typedef struct text16_font_reg_s 
{
	kgiu8 SR2,SR4;
	kgiu8 GR4,GR5,GR6;
} text16_font_reg;

static inline void text16_font_enter(text16_font_reg *regs) 
{
	IS_CRITICAL;

	ENTER_CRITICAL;

	regs->SR2 = CLIO_in8(SR2_PLANE_MASK);
	regs->SR4 = CLIO_in8(SR4_MEMORY_MODE);  /*  Save register state  */
	regs->GR4 = CLIO_in8(GR4_READ_MAP_SELECT);
	regs->GR5 = CLIO_in8(GR5_MODE);
	regs->GR6 = CLIO_in8(GR6_MISCELLANEOUS);

	CLIO_out8(SR2_PLANE_MASK, SR2_PLANE_2);	/* Font data goes to plane 2 */
	CLIO_out8(SR4_MEMORY_MODE, SR4_EXTENDED_MEMORY | SR4_ODD_EVEN);

	CLIO_out8(GR4_READ_MAP_SELECT, GR4_PLANE2); /* Read plane 2 */
	CLIO_out8(GR5_MODE, GR5_WRITE_MODE_0);
	CLIO_out8(GR6_MISCELLANEOUS, GR6_MEMORY_MAP_VGA | GR6_GRAPHICS_MODE);

	LEAVE_CRITICAL;
}

static void inline text16_font_leave(text16_font_reg *regs)
{
	IS_CRITICAL;

	ENTER_CRITICAL;

	CLIO_out8(SR2_PLANE_MASK, regs->SR2);
	CLIO_out8(SR4_MEMORY_MODE, regs->SR4);	/* Restore registers  */
	CLIO_out8(GR4_READ_MAP_SELECT, regs->GR4);
	CLIO_out8(GR5_MODE, regs->GR5);
	CLIO_out8(GR6_MISCELLANEOUS, regs->GR6);

	LEAVE_CRITICAL;
}

static inline void direct_text16_font_read(int chr, kgiu8 *charcell)
{
	memcpy(charcell, (kgiu8 *)(vga_mem_graph.base) + (chr << 5), 32);
}

static inline void direct_text16_font_write(int chr, kgiu8 *charcell)
{
	memcpy((kgiu8 *)(vga_mem_graph.base) + (chr << 5), charcell, 32);
}

static void text16_do_set_font(int no, int from, int to, struct kgi_font *font)
{
	text16_font_reg fontregs;
	int i;
	char *fbuf, *data;
	register int cnt = 32 - font->dy;

	if (font == NULL) {
		NOTICE("NULL font pointer given to set_font");
		return;
	}
	fbuf = (char *) vga_mem_graph.base;
	fbuf += (no * 0x2000) + (from << 5);
	data = font->data + font->dy * from;
	if (!data)
		return;
	text16_font_enter(&fontregs);
	/* NOTE: no printk()'s are allowed in here unless klogd is active! */

	for (i = from; i <= to; i++) {
		memmove(fbuf, data, font->dy);
		memset(fbuf + font->dy, 0, cnt);
		fbuf += 32;
		data += font->dy;
	};

	text16_font_leave(&fontregs);

#ifdef 0 /* CURSOR_HW */
	/* Set up cursor */
	outbCRT((inbCRT(0x0A) & ~CR0A_CURSOR_START_MASK)
		| ((font->base+1) & CR0A_CURSOR_START_MASK), 0x0A);
	outbCRT((inbCRT(0x0B) & ~CR0B_CURSOR_END_MASK)
		| ((font->base+2) & CR0B_CURSOR_END_MASK), 0x0B);
#endif
}

static void text16_set_font(struct kgi_display *dpy, struct kgi_font *font,
			    kgi_uint from, kgi_uint to)
{
	text16_do_set_font(0, from, to, font);
}

static void graphic_set_origin(struct kgi_display *dpy,
				kgi_uint dx, kgi_uint dy)
{
	int foo;
	int x = dx + dy*dpy->mode.request.virt.x;
	NOTICE1("Entering graphic_set_origin");
	/* Put x in bytes */
	switch(dpy->mode.request.graphtype) {
	case KGIGT_15BIT:
	case KGIGT_16BIT:	x <<= 1; break;
	case KGIGT_32BIT:	x <<= 2; break;
	default:	break;
	}
	DEBUG1("dx=%d dy=%d x = %d", dx,dy,x);
	#warning take care of tiling for set_origin -- ortalo
	CLIO_out8(CRD_SCREEN_START_ADDRESS_LOW, (x) & 0xFF);
	CLIO_out8(CRC_SCREEN_START_ADDRESS_HIGH, (x >> 8) & 0xFF);
	foo = CLIO_in8(CR1B_EXTENDED_DISPLAY_CONTROL);
	foo &=~(CR1B_SCREENA_ADDRESS_1718_MASK | CR1B_SCREENA_ADDRESS_16_MASK);
	foo |=  ( (x >> 16) & CR1B_SCREENA_ADDRESS_16_MASK ) |
		( (x >> (17 - 2)) & CR1B_SCREENA_ADDRESS_1718_MASK );
	CLIO_out8(CR1B_EXTENDED_DISPLAY_CONTROL, foo);
	foo = CLIO_in8(CR1D_SCREENA_START_EXTENSION);
	foo &= ~CR1D_SCREENA_ADDRESS_1920_MASK;
	foo |=  (x >> (19 - 3)) & CR1D_SCREENA_ADDRESS_1920_MASK;
	CLIO_out8(CR1D_SCREENA_START_EXTENSION, foo);

	NOTICE1("Leaving graphic_set_origin");
	return;
}
static void text16_set_origin(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy,chipset);
	IS_CRITICAL;

	register kgi_uint xpix = x / dpy->mode.request.dpp.x;
	register kgi_uint xdot = x % dpy->mode.request.dpp.x;
	register kgi_uint ypix = y / dpy->mode.request.dpp.y;
	register kgi_uint ydot = y % dpy->mode.request.dpp.y;
	register kgi_uint offset = xpix + ypix*dpy->fb.size.x;

	private->origin = offset;

	ENTER_CRITICAL;

		if (private->last_offset != offset) {

			CLIO_out8(CRC_SCREEN_START_ADDRESS_HIGH,
				(offset >> 8) & 0xFF);
			CLIO_out8(CRD_SCREEN_START_ADDRESS_LOW,
				 offset & 0xFF);
			private->last_offset = offset;
		}

		if (private->last_ydot != ydot) {

			CLIO_out8(CR8_SCREENA_PRESET_ROW_SCAN,
				 ydot & CR8_ROW_SCAN_MASK);

			/* NB: Byte pan is alway 0 (?) -- ortalo */
			private->last_ydot = ydot;
		}

		if (private->last_xdot != xdot) {

			register kgi_uint hpan;

			if (dpy->mode.request.dpp.x == 9) {

				hpan = (xdot == 0)
					    ? 0x08 : (xdot-1);
			} else {

				hpan = xdot;
			}
			/* This active wait should not be necessary
			** thanks to bit 1 ot CRT1A (see set_mode)
			*/
			#if 0
			wait_display_off(dpy);
			#endif

			CLIO_out8(AR13_PIXEL_PANNING,
				hpan & AR13_PIXEL_PANNING_MASK);

			private->last_xdot = xdot;
		}

	LEAVE_CRITICAL;
}

static void text16_set_splitline(struct kgi_display *dpy, int line)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy,chipset);
	register kgiu8 foo;

	NOTICE1("Entering text16_set_splitline");
	DEBUG1("split line=%d",line);
	if (private->splitline == line) {

		return;	/* Same as previous one */
	}

	CLIO_out8(CR18_LINE_COMPARE, line & 0xFF);

	foo = (line & 0x0100)
		? CLIO_in8(CR7_OVERFLOW) | CR7_LINE_COMPARE_8
		: CLIO_in8(CR7_OVERFLOW) & ~CR7_LINE_COMPARE_8;
	CLIO_out8(CR7_OVERFLOW, foo);

	foo = (line & 0x0200)
		? CLIO_in8(CR9_CHARACTER_CELL_HEIGHT) | CR9_LINE_COMPARE_9
		: CLIO_in8(CR9_CHARACTER_CELL_HEIGHT) & ~CR9_LINE_COMPARE_9;
	CLIO_out8(CR9_CHARACTER_CELL_HEIGHT, foo);

	private->splitline=line;
	NOTICE1("Leaving text16_set_splitline");
}

static void text16ShowCursor(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	struct chipset_private *private = KGIM_PRIVATE(dpy,chipset);
	IS_CRITICAL;
	register kgiu32 new = private->origin +
		(x / dpy->mode.request.dpp.x) +
		(y / dpy->mode.request.dpp.y) * dpy->fb.size.x;
/*
	#warning check fb_size for cursor -- ortalo
	new = (new < private->fb_size) ? new : new - private->fb_size;
	ASSERT(new < private->fb_size);
*/

	ENTER_CRITICAL;

		CLIO_out8(CRE_TEXT_CURSOR_LOCATION_HIGH, (new >> 8) & 0xFF);
		CLIO_out8(CRF_TEXT_CURSOR_LOCATION_LOW, new & 0xFF);

	LEAVE_CRITICAL;
}

static void text16HideCursor(struct kgi_display *dpy)
{
	IS_CRITICAL;
	ENTER_CRITICAL;

		CLIO_out8(CRE_TEXT_CURSOR_LOCATION_HIGH, 0xFF);
		CLIO_out8(CRF_TEXT_CURSOR_LOCATION_LOW, 0xFF);

	LEAVE_CRITICAL;
}

#define text16UndoCursor	text16HideCursor

static void text16ShowPointer(struct kgi_display *dpy, kgi_uint x, kgi_uint y)
{
	NOTICE2("Empty text16ShowPointer");
}

static void text16HidePointer(struct kgi_display *dpy)
{
	NOTICE2("Empty text16HidePointer");
}

#define text16UndoPointer	text16HidePointer
